"""Wrapper module for the Plotext _figure_class.

Designed to help type hint the methods in the class. See also the
accompanying pyi file for the types.
"""

from plotext._figure import _figure_class

__all__ = ["_figure_class"]
